/*
 * ATSAMD51WS2812B.c
 *
 * Created: 29.08.2018 18:31:47
 * Author : Thomas F
 */ 

#include <stdbool.h>

#include "sam.h"

#include "clocks.h"
#include "IO.h"
#include "misc.h"
#include "dma.h"

#include "HSV2RGB/HSV2RGB.h"
#include "WsDmaPort/WsDmaPort.h"





//    Used Pins
//
//      TxLed_PA27
//  WS_LOW_A4_PB08			
// WS_HIGH_A5_PB09			
// WS_DAT_D03_PB16			
// WS_DAT_D02_PB17			
//     NEOPIX_PB22				

// Used Peripherals
//
// SysTick 
// TC4
// DMA Channel 0-2
//



#define LED_COUNT 240
#define DMA_COUNT (LED_COUNT * 3 * 8)

static HSV_t HsvStripe[8][LED_COUNT];
static RGB_t RgbStripe[8][LED_COUNT];

__attribute__((aligned(4)))	// Because of word access in "ConvertRgbToDma8Channels()"
static uint8_t WsDmaBuffer[DMA_COUNT];

static uint8_t* WsRgbData[8] = { (uint8_t*)&RgbStripe[0], (uint8_t*)&RgbStripe[0], 0, 0, 0, 0, (uint8_t*)&RgbStripe[1], 0 };



// 100 Hz Timer ISR
static volatile _Bool Pulse10ms;

ISR(SysTick_Handler)
{
	Pulse10ms = true;
}




int main(void)
{
    // Initialize the SAM system
    SystemInit();
	SystemInitClock();

	
	// Enable Cortex M Cache Controller
	CMCC->CFG.bit.DCDIS = true;
	CMCC->CTRL.reg = CMCC_CTRL_CEN;


	// Set direction of general pins
	PORTA->DIRSET.reg = DEBUG_A0_PA02 | TxLed_PA27;

	
	// Initialize functions
	DmaInit();
 
	// Set direction for used WS pins
	PORTB->DIRSET.reg = WS_DAT_D03_PB16 | WS_DAT_D02_PB17 | NEOPIX_PB22;
	
	WsDmaPortInit(PORTB, 2, 0b01000011, WsDmaBuffer, sizeof(WsDmaBuffer), WsRgbData, false);


	// Prepare LED effect
	for (uint32_t i = 0; i < LED_COUNT; i++)
	{
		HsvStripe[0][i].H = i * -16;
		HsvStripe[0][i].S = 250;
		HsvStripe[0][i].V = 16;

		HsvStripe[1][i].H = i * -2;
		HsvStripe[1][i].S = 250;
		HsvStripe[1][i].V = 16;
	}

		
	// Timer 100 Hz
	__NVIC_SetPriority(SysTick_IRQn, 7);			// Low priority
	
	SysTick->LOAD  = SystemCoreClock / 100;			// Set reload register 100Hz
	SysTick->VAL   = 0;								// Load the SysTick Counter Value
	SysTick->CTRL  = SysTick_CTRL_CLKSOURCE_Msk		// Enable SysTick IRQ and SysTick Timer
				   | SysTick_CTRL_ENABLE_Msk
				   | SysTick_CTRL_TICKINT_Msk;
	
	
	// Main loop
    while (1) 
    {
		// Wait until last transfer is finished
		if (IsWsDmaPortTransferActiv())
		{
			Pulse10ms = false;
			continue;
		}

	
// Stopping the timer is not necessary, it's for debugging purposes
TC4->COUNT8.CTRLA.bit.ENABLE = false;
		
		
		// Start new transfer every 10ms (100Hz)
		if (Pulse10ms)
		{
			static uint32_t Counter1Hz;
			
			Pulse10ms = false;
			
			if (++Counter1Hz >= 50)
			{
				Counter1Hz = 0;
				PORTA->OUTTGL.reg = TxLed_PA27;
			}


			// Update LED effect
			for (uint32_t i = 0; i < LED_COUNT; i++)
			{
				HsvStripe[0][i].H += 1;
				HSV2RGB(&HsvStripe[0][i], &RgbStripe[0][i]);
				
				HsvStripe[1][i].H += 2;
				HSV2RGB(&HsvStripe[1][i], &RgbStripe[1][i]);
			}
	
			
			// Convert RGB data to DMA Port data
			ConvertRgbToDma8Channels();				


			// Start transfer
			WsDmaPortStartTransfer();

		}
    }
}









// For debugging purposes
void Fault_Handler(uint32_t number)
{
	CMCC->MCFG.reg = CMCC_MCFG_MODE_CYCLE_COUNT;
	CMCC->MEN.reg = CMCC_MEN_MENABLE;
	
	while(1)
	{	
		for (uint32_t i = 0; i < number; i++)
		{
			PORTA->OUTCLR.reg = ErrorLed_PAx;

			while(1)
			{
				if (CMCC->MSR.reg > (SystemCoreClock / 20))
				{
					CMCC->MCTRL.reg = CMCC_MCTRL_SWRST;
					break;
				}
			}

			PORTA->OUTSET.reg = ErrorLed_PAx;

			while(1)
			{
				if (CMCC->MSR.reg > (SystemCoreClock / 2))
				{
					CMCC->MCTRL.reg = CMCC_MCTRL_SWRST;
					break;
				}
			}
		}

		while(1)
		{
			if (CMCC->MSR.reg > (SystemCoreClock / 1))
			{
				CMCC->MCTRL.reg = CMCC_MCTRL_SWRST;
				break;
			}
		}

	}	
}


ISR(HardFault_Handler)	{ Fault_Handler(1); }

ISR(MemManage_Handler)	{ Fault_Handler(2); }

ISR(BusFault_Handler)	{ Fault_Handler(3); }

ISR(SVC_Handler)		{ Fault_Handler(4); }

ISR(DebugMon_Handler)	{ Fault_Handler(5); }

ISR(PendSV_Handler)		{ Fault_Handler(6); }

ISR(UsageFault_Handler)	{ Fault_Handler(7); }





